#include "M3.hpp"
using namespace std;

////////////////////////////////////// ParserM3 Macros /////////////////////////////////////

#define WriteValueHTML(Value){\
	ParserM3.Out << "<span class='varname'>" << #Value << "</span>: " << Value << "</br>";\
}

/////////////////////////////////////////////// M3 ///////////////////////////////////////////

M3::ParserMap M3::Parsers = M3::Initialize();

M3::ParserMap M3::Initialize()
{
	M3::ParserMap Parsers;
	Parsers[(uint32&)"RAHC"] = &M3::ParseCHAR;
	Parsers[(uint32&)"__8U"] = &M3::ParseU8;
	Parsers[(uint32&)"_61U"] = &M3::ParseU16;
	Parsers[(uint32&)"_23I"] = &M3::ParseI32;
	Parsers[(uint32&)"_23U"] = &M3::ParseU32;
	Parsers[(uint32&)"LAER"] = &M3::ParseREAL;
	Parsers[(uint32&)"2CEV"] = &M3::ParseVEC2;
	Parsers[(uint32&)"3CEV"] = &M3::ParseVEC3;
	Parsers[(uint32&)"4CEV"] = &M3::ParseQUAT;	// use quat to parse vec4 since they are essentialy the same
	Parsers[(uint32&)"TAUQ"] = &M3::ParseQUAT;
	Parsers[(uint32&)"SDNB"] = &M3::ParseBNDS;
	Parsers[(uint32&)"FERI"] = &M3::ParseIREF;
	Parsers[(uint32&)"33DM"] = &M3::ParseMD33;
	Parsers[(uint32&)"LDOM"] = &M3::ParseMODL;
	Parsers[(uint32&)"SQES"] = &M3::ParseSEQS;
	Parsers[(uint32&)"VEDS"] = &M3::ParseSDEV;
	Parsers[(uint32&)"V3DS"] = &M3::ParseSD3V;
	Parsers[(uint32&)"3RDS"] = &M3::ParseSDR3;
	Parsers[(uint32&)"Q4DS"] = &M3::ParseSD4Q;
	Parsers[(uint32&)"BMDS"] = &M3::ParseSDMB;
	Parsers[(uint32&)"_GTS"] = &M3::ParseSTG;
	Parsers[(uint32&)"_CTS"] = &M3::ParseSTC;
	Parsers[(uint32&)"_STS"] = &M3::ParseSTS;
	Parsers[(uint32&)"_VID"] = &M3::ParseDIV;
	Parsers[(uint32&)"NGER"] = &M3::ParseREGN;
	Parsers[(uint32&)"_TAB"] = &M3::ParseBAT;
	Parsers[(uint32&)"CESM"] = &M3::ParseMSEC;
	Parsers[(uint32&)"ENOB"] = &M3::ParseBONE;
	Parsers[(uint32&)"_TTA"] = &M3::ParseATT;
	Parsers[(uint32&)"TNVE"] = &M3::ParseEVNT;
	Parsers[(uint32&)"MTAM"] = &M3::ParseMATM;
	Parsers[(uint32&)"_TAM"] = &M3::ParseMAT;
	Parsers[(uint32&)"RYAL"] = &M3::ParseLAYR;
	return Parsers;
}

M3::~M3() 
{
	delete [] ReferenceTable;
}

void M3::ConvertM3toHTML(istream& In, ostream& Out)
{
	// write css link
	Out << "<link rel='stylesheet' href='ViewM3.css' type='text/css'/>";

	// parse in M3, render out HTML
	M3 ConverterM3(In, Out);
	ConverterM3.ConvertM3toHTML();
}

void M3::ConvertM3toHTML()
{
	Record Header;

	// read in the file header and then the reference table
	In.read((char*)&Header, sizeof(Header));
	ReferenceTable = new Record [Header.Size];
	In.seekg(Header.Address);
	In.read((char*)ReferenceTable, Header.Size * sizeof(Record));
	
	// for each record in the refernce table - parse and render it
	for(uint32 i = 0; i < Header.Size; ++i){
		Parse(ReferenceTable[i]);
	}

	// write the table of contents
	WriteSectionStartHTML(Header);
	for(uint32 i = 0; i < Header.Size; ++i){
		WriteRecordHTML(i, ReferenceTable[i]);
	}
	WriteSectionEndHTML();
}

void M3::Parse(const Record& Record)
{
	ParserMap::iterator I;

	WriteSectionStartHTML(Record);	// render the <div class='section'>
	In.seekg(Record.Address);		// move to the file section
	if((I = Parsers.find(Record.Type.Value)) != Parsers.end()){	// lookup parser
		(this->*I->second)(Record);	// parse and render
	}
	WriteSectionEndHTML(); // end with </div>
}

void M3::ParseCHAR(const Record& Record)
{
	char* String = new char [Record.Size];
	In.read(String, Record.Size);
	Out << "<div>'" << String << "'</div>";
}

void M3::ParseU8(const Record& Record)
{
	ParseArray<U8>(Record);
}

void M3::ParseU16(const Record& Record)
{
	ParseArray<U16>(Record);
}

void M3::ParseI32(const Record& Record)
{
	ParseArray<I32>(Record);
}

void M3::ParseU32(const Record& Record)
{
	ParseArray<U32>(Record);
}

void M3::ParseREAL(const Record& Record)
{
	ParseArray<REAL>(Record);
}

void M3::ParseVEC2(const Record& Record)
{
	ParseArray<VEC2>(Record);
}

void M3::ParseVEC3(const Record& Record)
{
	ParseArray<VEC3>(Record);
}

void M3::ParseQUAT(const Record& Record)
{
	ParseArray<QUAT>(Record);
}

void M3::ParseIREF(const Record& Record)
{
	ParseArray<IREF>(Record);
}

void M3::ParseBNDS(const Record& Record)
{
	ParseArray<BNDS>(Record);
}

void M3::ParseMD33(const Record& Record)
{
	ParseSection<MD33>(Record);
}

void M3::ParseMODL(const Record& Record)
{
	if(Record.Flags == 23){
		ParseSection<MODL23>(Record);
	}
	else{
		ParseSection<MODL20>(Record);
	}
}

void M3::ParseSEQS(const Record& Record)
{
	ParseSections<SEQS>(Record);
}

void M3::ParseSTG(const Record& Record)
{
	ParseSections<STG>(Record);
}

void M3::ParseSTC(const Record& Record)
{
	ParseSections<STC>(Record);
}

void M3::ParseSTS(const Record& Record)
{
	ParseSections<STS>(Record);
}

void M3::ParseSDEV(const Record& Record)
{
	ParseSections<SDEV>(Record);
}

void M3::ParseSD3V(const Record& Record)
{
	ParseSections<SD3V>(Record);
}

void M3::ParseSDR3(const Record& Record)
{
	ParseSections<SDR3>(Record);
}

void M3::ParseSD4Q(const Record& Record)
{
	ParseSections<SD4Q>(Record);
}

void M3::ParseSDMB(const Record& Record)
{
	ParseSections<SDMB>(Record);
}

void M3::ParseBONE(const Record& Record)
{
	ParseSections<BONE>(Record);
}

void M3::ParseATT(const Record& Record)
{
	ParseSections<ATT>(Record);
}

void M3::ParseEVNT(const Record& Record)
{
	ParseSections<EVNT>(Record);
}

void M3::ParseDIV(const Record& Record)
{
	ParseSections<DIV>(Record);
}

void M3::ParseREGN(const Record& Record)
{
	ParseSections<REGN>(Record);
}

void M3::ParseBAT(const Record& Record)
{
	ParseSections<BAT>(Record);
}

void M3::ParseMSEC(const Record& Record)
{
	ParseSections<MSEC>(Record);
}

void M3::ParseMATM(const Record& Record)
{
	ParseSections<MATM>(Record);
}

void M3::ParseMAT(const Record& Record)
{
	ParseSections<MAT>(Record);
}

void M3::ParseLAYR(const Record& Record)
{
	ParseSections<LAYR>(Record);
}

//////////////////////////////////////////// WriteHTML /////////////////////////////////////

void M3::WriteSectionStartHTML(const Record& Record)
{
	Out 
		<< "<div class='section'>\n\t<a name='a_" << hex << Record.Address 
		<< "'></a>\n\t<h3>" << dec << Record.Type
		<< "-" << Record.Flags
		<< "</h3> [Address: <b>0x" << hex << Record.Address
		<< "</b>, Size: <b>" << dec << Record.Size 
		<< "</b>]<br/>\n";
}

void M3::WriteSectionEndHTML()
{
	Out << "</div>\n";
}

void M3::WriteRecordHTML(uint32 Index, const Record& Record)
{
	Out 
		<< "<div><a name='i_" << Index 
		<< "'></a><b>#" << Index 
		<< "</b> " << Record.Type
		<< "-" << Record.Flags 
		<< " ["	<< Record.Size << "] <a href='#a_" 
		<< hex << Record.Address << "'> 0x"
		<< Record.Address << dec << "</a></div>\n";
}

void M3::WriteReferenceHTML(const Reference& Reference, const char* Name)
{
	Record& Record(ReferenceTable[Reference.Index]);
	
	if(Name){
		Out << "<span class='varname'>" << Name << "</span>: ";
	}
	Out
		<< "[" << Record.Type
		<< "-" << Record.Flags
		<< "] <b>" << Reference.Size
		<< "</b> <a href='#a_" << hex << Record.Address
		<< "'>@0x" << Record.Address << dec << "</a>";

	if(Record.Type.Value == (uint32&)"RAHC"){
		char* String = new char[Record.Size];
		In.seekg(Record.Address);
		In.read(String, Record.Size);
		Out << " => '" << String << "'</br>\n";
		delete [] String;
	}
	else{
		Out << "</br>\n";
	}
}

void M3::WriteReferenceHTML(const AnimationReference& Reference)
{
	Out
		<< "<span class='varname'>Animation</span>[Flags: " << Reference.Flags
		<< ", AnimFlags: " << Reference.AnimationFlags
		//<< "] </b> <a href='#id_" << hex << Reference.AnimationID
		//<< "'>#" << Reference.AnimationID << dec << "</a></br>\n";
		<< ", ID: <b>"<< Reference.AnimationID << dec << "</b>]</br>\n";
}

void M3::WriteHexDumpHTML(uint8 Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint8") 
		<< "</span><span class='unknown'>: " << hex;
	WriteHex(Out, Unknown);
	Out << " " << dec << "</span><br/>\n";
}

void M3::WriteHexDumpHTML(int16 Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "int16")
		<< "</span><span class='unknown'>: " << hex;
	WriteHex(Out, Unknown);
	Out << " " << dec << "</span><br/>\n";
}

void M3::WriteHexDumpHTML(uint16 Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint16")
		<< "</span><span class='unknown'>: " << hex;
	WriteHex(Out, Unknown);
	Out << " " << dec << "</span><br/>\n";
}

void M3::WriteHexDumpHTML(int32 Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "int32")
		<< "</span><span class='unknown'>: " << hex;
	WriteHex(Out, Unknown);
	Out << " " << dec << "</span><br/>\n";
}

void M3::WriteHexDumpHTML(uint32 Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint32")
		<< "</span><span class='unknown'>: " << hex;
	WriteHex(Out, Unknown);
	Out << " " << dec << "</span><br/>\n";
}

void M3::WriteHexDumpHTML(float Unknown, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "float") 
		<< "</span><span class='unknown'>: " << Unknown
		<< "</span><br/>";
}

void M3::WriteHexDumpHTML(const uint8* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint8") 
		<< "</span><span class='unknown'>: " << hex;
	for(uint i = 0; i < Size; ++i){
		WriteHex(Out, Unknown[i]);
	}
	Out << dec << "</span><br/>";
}

void M3::WriteHexDumpHTML(const int16* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "int16")
		<< "</span><span class='unknown'>: " << hex;
	for(uint i = 0; i < Size; ++i){
		WriteHex(Out, Unknown[i]);
		Out << " ";
	}
	Out << dec << "</span><br/>";
}

void M3::WriteHexDumpHTML(const uint16* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint16")
		<< "</span><span class='unknown'>: " << hex;
	for(uint i = 0; i < Size; ++i){
		WriteHex(Out, Unknown[i]);
		Out << " ";
	}
	Out << dec << "</span><br/>";
}

void M3::WriteHexDumpHTML(const int32* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "int32")
		<< "</span><span class='unknown'>: " << hex;
	for(uint i = 0; i < Size; ++i){
		WriteHex(Out, Unknown[i]);
		Out << " ";
	}
	Out << dec << "</span><br/>";
}

void M3::WriteHexDumpHTML(const uint32* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "uint32")
		<< "</span><span class='unknown'>: " << hex;
	for(uint i = 0; i < Size; ++i){
		WriteHex(Out, Unknown[i]);
		Out << " ";
	}
	Out << dec << "</span><br/>";
}

void M3::WriteHexDumpHTML(const float* Unknown, uint Size, const char* Name)
{
	Out 
		<< "<span class='varname unknown'>" << (Name ? Name : "float")
		<< "</span><span class='unknown'>: ";
	for(uint i = 0; i < Size; ++i){
		Out << Unknown[i] << " ";
	}
	Out << "</span><br/>";
}

void M3::WriteVerticesHTML(const Reference& Reference, uint32 Flags)
{
	Record& Record(ReferenceTable[Reference.Index]);
	uint NumberOfVertices;

	Out << "<div><a name='vertices'></a><h3>Vertices</h3> => \n";
	if(Flags & 0x40000){
		NumberOfVertices = Record.Size/sizeof(Vertex28);
		Vertex28* Vertices = new Vertex28 [NumberOfVertices];
		In.seekg(Record.Address);
		In.read((char*)Vertices, Record.Size);

		Out << "Position Normal Tangent UV UV2? Weights[4] Indices[4] <br/>";
		for(uint i = 0; i < NumberOfVertices; ++i){
			Out << Vertices[i] << "<br/>\n";
		}
	}
	else{
		NumberOfVertices = Record.Size/sizeof(Vertex24);
		Vertex24* Vertices = new Vertex24 [NumberOfVertices];
		In.seekg(Record.Address);
		In.read((char*)Vertices, Record.Size);

		Out << "Position Normal Tangent UV Weights[4] Indices[4]<br/>";
		for(uint i = 0; i < NumberOfVertices; ++i){
			Out << Vertices[i] << "<br/>\n";;
		}
	}
}

void M3::MD33::WriteHTML(M3& ParserM3) const
{
	ParserM3.Out
		<< "<b>Magic:</b> " << Type
		<< "</br>\n<b>Type:</b> " << Flags
		<< "</br>\n<b>Table of Contents:</b> <a href='#a_" << hex << Address 
		<< "'>@0x" << Address << dec << "</a></br>\n"
		<< "<b>Number of Records in TOC:</b> " << Record::Size 
		<< "</br> Root = ";
	ParserM3.WriteReferenceHTML((M3::Reference&)Flags);
}

void M3::MODL20::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(MaxFilenameCHAR, "MaxFilename");
	WriteValueHTML(Version);
	ParserM3.WriteReferenceHTML(SequencesSEQS, "Sequences");
	ParserM3.WriteReferenceHTML(SequencesSTC, "Sequence Table C");
	ParserM3.WriteReferenceHTML(SequencesSTG, "Sequence Table G");
	ParserM3.WriteHexDumpHTML(Unknown0, 3);
	ParserM3.WriteReferenceHTML(SequencesSTS, "Sequence Table S");
	ParserM3.WriteReferenceHTML(BonesBONE, "Bones");
	WriteValueHTML(Unknown1);
	WriteValueHTML(Flags);
	ParserM3.WriteReferenceHTML(VerticesU8, "Vertices");
	ParserM3.Out << "<a href='#vertices'>View Vertices</a><br/>";
	ParserM3.WriteReferenceHTML(GeometryDIV, "Geometry");
	ParserM3.WriteReferenceHTML(BoneLookupU16, "Bone Lookup");
	WriteValueHTML(Bounds);
	ParserM3.WriteHexDumpHTML(Unknown2, 13);
	ParserM3.WriteReferenceHTML(AttachmentATT, "Attachments");
	ParserM3.WriteReferenceHTML(AttachmentLookupU16, "Attachment Lookup");
	ParserM3.WriteReferenceHTML(Lights, "Lights");
	ParserM3.WriteReferenceHTML(Cameras, "Cameras");
	ParserM3.WriteReferenceHTML(A_U16);
	ParserM3.WriteReferenceHTML(MaterialGroupsMATM, "Material Table");
	ParserM3.WriteReferenceHTML(MaterialMAT, "Material");
	ParserM3.WriteReferenceHTML(DIS, "DIS");
	ParserM3.WriteReferenceHTML(CMP, "CMP");
	ParserM3.WriteReferenceHTML(TER, "TER"); 
	ParserM3.WriteReferenceHTML(VOL, "VOL");
	ParserM3.WriteHexDumpHTML(Unknown3, 2);
	ParserM3.WriteReferenceHTML(CREP, "CREP");
	ParserM3.WriteReferenceHTML(PAR, "PAR");
	ParserM3.WriteReferenceHTML(PARC, "PARC");
	ParserM3.WriteReferenceHTML(RIB, "RIB");
	ParserM3.WriteReferenceHTML(PROJ, "PROJ");
	ParserM3.WriteReferenceHTML(FOR, "FOR");
	ParserM3.WriteReferenceHTML(WRP, "WRP");
	ParserM3.WriteHexDumpHTML(Unknown4, 2);
	ParserM3.WriteReferenceHTML(PHRB, "PHRB");
	ParserM3.WriteHexDumpHTML(Unknown5, 6);
	ParserM3.WriteReferenceHTML(IKJT, "IKJT");
	ParserM3.WriteHexDumpHTML(Unknown6, 2);
	ParserM3.WriteReferenceHTML(PATU, "PATU");
	ParserM3.WriteReferenceHTML(TRGD), "TRGD";
	ParserM3.WriteReferenceHTML(BonesIREF, "Bone Matrices");
	ParserM3.WriteReferenceHTML(B_I32);
	WriteValueHTML(Matrix);
	WriteValueHTML(Bounds1);
	ParserM3.WriteReferenceHTML(SSGS, "SSGS");
	ParserM3.WriteReferenceHTML(ATVL, "ATVL");
	ParserM3.WriteReferenceHTML(BBSC, "BBSC");
	ParserM3.WriteReferenceHTML(TDM, "TDM");
	ParserM3.WriteHexDumpHTML(Unknown7, 3);

	// write vertices
	ParserM3.WriteVerticesHTML(VerticesU8, Flags);
}

void M3::MODL23::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(MaxFilenameCHAR, "MaxFilename");
	WriteValueHTML(Version);
	ParserM3.WriteReferenceHTML(SequencesSEQS, "Sequences");
	ParserM3.WriteReferenceHTML(SequencesSTC, "Sequence Table C");
	ParserM3.WriteReferenceHTML(SequencesSTG, "Sequence Table G");
	ParserM3.WriteHexDumpHTML(Unknown0, 3);
	ParserM3.WriteReferenceHTML(SequencesSTS, "Sequence Table S");
	ParserM3.WriteReferenceHTML(BonesBONE, "Bones");
	WriteValueHTML(Unknown1);
	WriteValueHTML(Flags);
	ParserM3.WriteReferenceHTML(VerticesU8, "Vertices");
	ParserM3.Out << "<a href='#vertices'>View Vertices</a><br/>";
	ParserM3.WriteReferenceHTML(GeometryDIV, "Geometry");
	ParserM3.WriteReferenceHTML(BoneLookupU16, "Bone Lookup");
	WriteValueHTML(Bounds);
	ParserM3.WriteHexDumpHTML(Unknown2, 13);
	ParserM3.WriteReferenceHTML(AttachmentATT, "Attachments");
	ParserM3.WriteReferenceHTML(AttachmentLookupU16, "Attachment Lookup");
	ParserM3.WriteReferenceHTML(Lights, "Lights");
	ParserM3.WriteReferenceHTML(SHBX, "SHBX");
	ParserM3.WriteReferenceHTML(Cameras, "Cameras");
	ParserM3.WriteReferenceHTML(A_U16);
	ParserM3.WriteReferenceHTML(MaterialGroupsMATM, "Material Table");
	ParserM3.WriteReferenceHTML(MaterialMAT, "Material");
	ParserM3.WriteReferenceHTML(DIS, "DIS");
	ParserM3.WriteReferenceHTML(CMP, "CMP");
	ParserM3.WriteReferenceHTML(TER, "TER"); 
	ParserM3.WriteReferenceHTML(VOL, "VOL");
	ParserM3.WriteHexDumpHTML(Unknown3, 2);
	ParserM3.WriteReferenceHTML(CREP, "CREP");
	ParserM3.WriteReferenceHTML(PAR, "PAR");
	ParserM3.WriteReferenceHTML(PARC, "PARC");
	ParserM3.WriteReferenceHTML(RIB, "RIB");
	ParserM3.WriteReferenceHTML(PROJ, "PROJ");
	ParserM3.WriteReferenceHTML(FOR, "FOR");
	ParserM3.WriteReferenceHTML(WRP, "WRP");
	ParserM3.WriteHexDumpHTML(Unknown4, 2);
	ParserM3.WriteReferenceHTML(PHRB, "PHRB");
	ParserM3.WriteHexDumpHTML(Unknown5, 6);
	ParserM3.WriteReferenceHTML(IKJT, "IKJT");
	ParserM3.WriteHexDumpHTML(Unknown6, 2);
	ParserM3.WriteReferenceHTML(PATU, "PATU");
	ParserM3.WriteReferenceHTML(TRGD), "TRGD";
	ParserM3.WriteReferenceHTML(BonesIREF, "Bone Matrices");
	ParserM3.WriteReferenceHTML(B_I32);
	WriteValueHTML(Matrix);
	WriteValueHTML(Bounds1);
	ParserM3.WriteReferenceHTML(SSGS, "SSGS");
	ParserM3.WriteReferenceHTML(ATVL, "ATVL");
	ParserM3.WriteReferenceHTML(C_U16);
	ParserM3.WriteReferenceHTML(D_U16);
	ParserM3.WriteReferenceHTML(BBSC, "BBSC");
	ParserM3.WriteReferenceHTML(TDM, "TDM");
	ParserM3.WriteHexDumpHTML(Unknown7, 3);

	// write vertices
	ParserM3.WriteVerticesHTML(VerticesU8, Flags);
}

void M3::SEQS::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteHexDumpHTML(d1, 2);
	ParserM3.WriteReferenceHTML(SequenceNameCHAR);
	WriteValueHTML(d2);
	WriteValueHTML(TimeLength);
	WriteValueHTML(d3);
	WriteValueHTML(Flags);
	ParserM3.WriteHexDumpHTML(d4, 5);
	WriteValueHTML(Bounds);
	ParserM3.WriteHexDumpHTML(d5, 2);
}

void M3::STG::WriteHTML(M3& ParserM3) const
{
    ParserM3.WriteReferenceHTML(SequenceNameCHAR);
    ParserM3.WriteReferenceHTML(SequenceIDU32);
}

void M3::STC::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(SequenceNameCHAR);
	ParserM3.WriteHexDumpHTML(d1);
	WriteValueHTML(SequenceIndex);
	WriteValueHTML(SequenceIndex1);
	ParserM3.WriteReferenceHTML(AnimationIDsU32);
	ParserM3.WriteReferenceHTML(AnimationLookupU32);
	ParserM3.WriteHexDumpHTML(d2);
	for(uint i = 0; i < 13; ++i){
		ParserM3.WriteReferenceHTML(SequenceData[i]);
	}
}

void M3::STS::WriteHTML(M3& ParserM3) const
{
    ParserM3.WriteReferenceHTML(AnimationIDsU32);
    ParserM3.WriteHexDumpHTML(d1, 3);
    WriteValueHTML(s1);
    WriteValueHTML(s2);
}

void M3::SD::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(TimelineI32);
	WriteValueHTML(Flags);
	WriteValueHTML(TimeLength);
};

void M3::SDEV::WriteHTML(M3& ParserM3) const
{
	SD::WriteHTML(ParserM3);
	ParserM3.WriteReferenceHTML(DataEVNT);
}

void M3::SD3V::WriteHTML(M3& ParserM3) const
{
	SD::WriteHTML(ParserM3);
	ParserM3.WriteReferenceHTML(DataVEC3);
}

void M3::SDR3::WriteHTML(M3& ParserM3) const
{
	SD::WriteHTML(ParserM3);
	ParserM3.WriteReferenceHTML(DataREAL);
}

void M3::SD4Q::WriteHTML(M3& ParserM3) const
{
	SD::WriteHTML(ParserM3);
	ParserM3.WriteReferenceHTML(DataQUAT);
}

void M3::SDMB::WriteHTML(M3& ParserM3) const
{
	SD::WriteHTML(ParserM3);
	ParserM3.WriteReferenceHTML(DataBNDS);
}

void M3::DIV::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(FacesU16);
	ParserM3.WriteReferenceHTML(RegionsREGN);
	ParserM3.WriteReferenceHTML(MaterialsBAT);
	ParserM3.WriteReferenceHTML(MSEC);
	WriteValueHTML(unk);
}

void M3::REGN::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(d1);
	WriteValueHTML(VertexStart);
	WriteValueHTML(VertexCount);
	WriteValueHTML(IndexStart);
	WriteValueHTML(IndexCount);
	WriteValueHTML(MaxBoneSize);
	WriteValueHTML(BoneStart);
	WriteValueHTML(BoneCount);
	ParserM3.WriteHexDumpHTML(s1, 3);
}

void M3::BAT::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(d1); 			// always 0?
	WriteValueHTML(RegionIndex); 	// associated REGN
	ParserM3.WriteHexDumpHTML(s1, 2);
	WriteValueHTML(MaterialIndex); 	// index into MATM table
	WriteValueHTML(s2); 			// usually -1
}

void M3::MSEC::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(d1); 				//always 0?
	ParserM3.WriteReferenceHTML(BoundsAnimation); 		//Bounding box animation ref in STC
	WriteValueHTML(ext1[0]);
	WriteValueHTML(ext1[1]);
	ParserM3.WriteHexDumpHTML(d2, 8);
}

void M3::BONE::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(d1);
	ParserM3.WriteReferenceHTML(BoneNameCHAR);
	WriteValueHTML(Flags);
	WriteValueHTML(ParentIndex);
	WriteValueHTML(s1);
	ParserM3.WriteReferenceHTML(TranslationAnimation);
	WriteValueHTML(Translation);
	ParserM3.WriteHexDumpHTML(d2, 4);
	ParserM3.WriteReferenceHTML(RotationAnimation);
	WriteValueHTML(Rotation);
	ParserM3.WriteHexDumpHTML(d3, 3);
	WriteValueHTML(f1);
	WriteValueHTML(d4);
	ParserM3.WriteReferenceHTML(ScaleAnimation);
	WriteValueHTML(Scale);
	WriteValueHTML(v1);
	ParserM3.WriteHexDumpHTML(d5, 6);
}

void M3::MATM::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(Size);
	WriteValueHTML(MaterialIndex);
}

void M3::MAT::WriteHTML(M3& ParserM3) const
{
    ParserM3.WriteReferenceHTML(MaterialNameCHAR);
    ParserM3.WriteHexDumpHTML(d1, 8);
    WriteValueHTML(x);
	WriteValueHTML(y);
	for(uint32 i = 0; i < 13; ++i){
    	ParserM3.WriteReferenceHTML(Layers[i]);
	}
    ParserM3.WriteHexDumpHTML(d2, 15);
}

void M3::LAYR::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteHexDumpHTML(d1);
	ParserM3.WriteReferenceHTML(LayerNameCHAR);
	ParserM3.WriteReferenceHTML(ar1);
	ParserM3.WriteHexDumpHTML(d2, 6); //d2[5] determines team colour
	ParserM3.WriteReferenceHTML(ar2);
	ParserM3.WriteHexDumpHTML(f1, 2);
	ParserM3.WriteHexDumpHTML(d3);
	ParserM3.WriteReferenceHTML(ar3);
	ParserM3.WriteHexDumpHTML(d4, 4);
	ParserM3.WriteHexDumpHTML(d5);
	ParserM3.WriteHexDumpHTML(d6, 5);
	ParserM3.WriteReferenceHTML(ar4);
	ParserM3.WriteHexDumpHTML(d7, 3);
	ParserM3.WriteReferenceHTML(ar5);
	ParserM3.WriteHexDumpHTML(d8, 5);
	ParserM3.WriteReferenceHTML(ar6);
	ParserM3.WriteHexDumpHTML(d9, 2);
	ParserM3.WriteReferenceHTML(ar7);
	ParserM3.WriteHexDumpHTML(f2, 2);
	ParserM3.WriteHexDumpHTML(d10, 3);
	ParserM3.WriteReferenceHTML(ar8);
	ParserM3.WriteHexDumpHTML(f3);
	ParserM3.WriteHexDumpHTML(d11);
	ParserM3.WriteHexDumpHTML(f4);
	ParserM3.WriteHexDumpHTML(d12, 4);
	ParserM3.WriteReferenceHTML(ar9);
	ParserM3.WriteHexDumpHTML(f5, 4);
	ParserM3.WriteHexDumpHTML(d13);
	ParserM3.WriteReferenceHTML(ar10);
	ParserM3.WriteHexDumpHTML(d14, 3);
	ParserM3.WriteReferenceHTML(ar11);
	ParserM3.WriteHexDumpHTML(f6, 2);
	ParserM3.WriteHexDumpHTML(d15);
	ParserM3.WriteReferenceHTML(ar12);
	ParserM3.WriteHexDumpHTML(f7, 2);
	ParserM3.WriteHexDumpHTML(d16);
	ParserM3.WriteHexDumpHTML(d17);
	ParserM3.WriteHexDumpHTML(d18);
	ParserM3.WriteHexDumpHTML(f8);
	ParserM3.WriteHexDumpHTML(d19);
	ParserM3.WriteHexDumpHTML(f9, 2);
}

void M3::CAM::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(dl);
	ParserM3.WriteReferenceHTML(CameraNameCHAR);
	WriteValueHTML(flags1);
	WriteValueHTML(flags2);
}

void M3::EVNT::WriteHTML(M3& ParserM3) const
{
	ParserM3.WriteReferenceHTML(EventNameCHAR);
	WriteValueHTML(s1[4]);
	WriteValueHTML(Matrix);
	WriteValueHTML(d1 [4]);
}

void M3::ATT::WriteHTML(M3& ParserM3) const
{
	WriteValueHTML(d1);
	ParserM3.WriteReferenceHTML(AttachmentNameCHAR);
	WriteValueHTML(BoneIndex);
}

/////////////////////////////////// OSTREAM OPERATORS /////////////////////////

ostream& operator << (ostream& Out, const M3::VEC2& Vector)
{
	return Out
		<< "(" << Vector.x 
		<< ", " << Vector.y
		<< ")";
}

ostream& operator << (ostream& Out, const M3::VEC3& Vector)
{
	return Out
		<< "(" << Vector.x 
		<< ", " << Vector.y
		<< ", " << Vector.z 
		<< ")";
}

ostream& operator << (ostream& Out, const M3::QUAT& Quaternion)
{
	return Out 
		<< "(" << Quaternion.w 
		<< ", " << Quaternion.x 
		<< ", " << Quaternion.y
		<< ", " << Quaternion.z 
		<< ")";
}

ostream& operator << (ostream& Out, const M3::BNDS& Bounds)
{
	return Out
		<< "[" << Bounds.v1 
		<< " - " << Bounds.v2 
		<< ", " << Bounds.Radius
		<< "]";
}

ostream& operator << (ostream& Out, const M3::IREF& Matrix)
{
	return Out
		<< "[(" << Matrix[0][0] << ", " << Matrix[0][1] << ", " << Matrix[0][2] << ", " << Matrix[0][3] 
		<< "), (" << Matrix[1][0] << ", " << Matrix[1][1] << ", " << Matrix[1][2] << ", " << Matrix[1][3] 
		<< "), (" << Matrix[2][0] << ", " << Matrix[2][1] << ", " << Matrix[2][2] << ", " << Matrix[2][3] 
		<< "), (" << Matrix[3][0] << ", " << Matrix[3][1] << ", " << Matrix[3][2] << ", " << Matrix[3][3] 
		<< ")]";
}

ostream& operator << (ostream& Out, const M3::Vertex24& Vertex)
{
	Vector3f Normal, Tangent;
	Vector4f Weights;
	Vector2f UV;
	
	Normal.x = (2*(float)Vertex.Normal[0]/255 - 1);
	Normal.y = (2*(float)Vertex.Normal[1]/255 - 1);
	Normal.z = (2*(float)Vertex.Normal[2]/255 - 1);
	if(Vertex.Normal[3]){
		Normal.x /= Vertex.Normal[3];
		Normal.y /= Vertex.Normal[3];
		Normal.z /= Vertex.Normal[3];
	}

	Tangent.x = (2*(float)Vertex.Tangent[0]/255 - 1);
	Tangent.y = (2*(float)Vertex.Tangent[1]/255 - 1);
	Tangent.z = (2*(float)Vertex.Tangent[2]/255 - 1);
	if(Vertex.Tangent[3]){
		Tangent.x /= Vertex.Tangent[3];
		Tangent.y /= Vertex.Tangent[3];
		Tangent.z /= Vertex.Tangent[3];
	}

	Weights.w = (float)Vertex.BoneWeights[0]/255;
	Weights.x = (float)Vertex.BoneWeights[1]/255;
	Weights.y = (float)Vertex.BoneWeights[2]/255;
	Weights.z = (float)Vertex.BoneWeights[3]/255;

	UV.x = (float)Vertex.UV[0]/2048;
	UV.y = (float)Vertex.UV[1]/2048;

	return Out 
		<< Vertex.Position
		<< " " << Normal 
		<< " " << Tangent
		<< " " << UV
		<< " " << Weights
		<< " [" << (uint16)Vertex.BoneIndices[0]
		<< ", " << (uint16)Vertex.BoneIndices[1]
		<< ", " << (uint16)Vertex.BoneIndices[2]
		<< ", " << (uint16)Vertex.BoneIndices[3]
		<< "]";
}

ostream& operator << (ostream& Out, const M3::Vertex28& Vertex)
{
	Vector3f Normal, Tangent;
	Vector4f Weights;
	Vector2f UV, Unknown;

	Normal.x = (2*(float)Vertex.Normal[0]/255 - 1);
	Normal.y = (2*(float)Vertex.Normal[1]/255 - 1);
	Normal.z = (2*(float)Vertex.Normal[2]/255 - 1);
	if(Vertex.Normal[3]){
		Normal.x /= Vertex.Normal[3];
		Normal.y /= Vertex.Normal[3];
		Normal.z /= Vertex.Normal[3];
	}

	Tangent.x = (2*(float)Vertex.Tangent[0]/255 - 1);
	Tangent.y = (2*(float)Vertex.Tangent[1]/255 - 1);
	Tangent.z = (2*(float)Vertex.Tangent[2]/255 - 1);
	if(Vertex.Tangent[3]){
		Tangent.x /= Vertex.Tangent[3];
		Tangent.y /= Vertex.Tangent[3];
		Tangent.z /= Vertex.Tangent[3];
	}

	Weights.w = (float)Vertex.BoneWeights[0]/255;
	Weights.x = (float)Vertex.BoneWeights[1]/255;
	Weights.y = (float)Vertex.BoneWeights[2]/255;
	Weights.z = (float)Vertex.BoneWeights[3]/255;

	UV.x = (float)Vertex.UV[0]/2048;
	UV.y = (float)Vertex.UV[1]/2048;

	Unknown.x = (float)Vertex.s1[0]/2048;
	Unknown.y = (float)Vertex.s1[1]/2048;

	return Out 
		<< Vertex.Position
		<< " " << Normal 
		<< " " << Tangent
		<< " " << UV
		<< " " << Unknown
		<< " " << Weights
		<< " [" << (uint16)Vertex.BoneIndices[0]
		<< ", " << (uint16)Vertex.BoneIndices[1]
		<< ", " << (uint16)Vertex.BoneIndices[2]
		<< ", " << (uint16)Vertex.BoneIndices[3]
		<< "]";
}